# version 2022.1
[CmdLetBinding()]
param(
	[string]$username,
	[string]$password,
	[Switch]$logToConsole
)

try {
	cd $PSScriptRoot
	$logFileName = 'attestation.{0}.log' -f (get-date -Format 'yyyy-MM')
	. ./utils.ps1 -logFileName $logFileName -logToConsole:$logToConsole
	$api = [mm]::getAPI($username, $password)
} catch {
	Write-Output ('Got exception while initializing: ' + $_.Exception.GetBaseException().Message)
	exit 1
}

. ./workflow_common.ps1

$settings = $Global:mmConfig.attestation
$subjects = $Global:mmConfig.subjects
$expiryWarning1Days = ?? $settings.expiryWarning1_Days 21
$expiryWarning2Days = ?? $settings.expiryWarning2_Days 14
$expiryWarning3Days = ?? $settings.expiryWarning3_Days 7
$expiryWarning4Days = ?? $settings.expiryWarning4_Days 0
$expiredGracePeriodDays = ?? $settings.expiredGracePeriodDays 7
$today = get-date

enum Templates {
	aboutToExpire
    expired
    deleted
    deleteFlag
}

function getTemplateInfo([Templates]$template){
    switch ($template){
        aboutToExpire {
            return @{
                file = "lifecycle-expiry-warning.htm"
                subject = (?? $subjects['lifecycle-expiry-warning'] "Review these external DNS entries that are about to expire")
            }
        }
        expired {
            return @{
                file = "lifecycle-expired-warning.htm"
                subject = (?? $subjects['lifecycle-expired-warning'] "External DNS entry will be deleted")
            }
        }
        deleted {
            return @{
                file = "lifecycle-expired-deleted.htm"
                subject = (?? $subjects['lifecycle-expired-deleted'] "External DNS entry has been deleted")
                cc = $settings.deletedEmailCC
            }
        }
        deleteFlag {
            return @{
                file = 'lifecycle-flagged-deletion.htm'
                subject = (?? $subjects['lifecycle-flagged-deletion'] "External DNS entry has been deleted")
                cc = $settings.deletedEmailCC
            }
        }
    }
}

function sendEmailNotification([hashtable[]]$records, [Templates]$template){
   
    function getTeamEmails(){
        $teamsDLs = @{}
        foreach ($record in $records){
            if ($record.TeamDL -and !$teamsDLs.ContainsKey($record.TeamDL)) {
                $teamsDLs[$record.TeamDL]=[adUtils]::getGroup($record.TeamDL).mail
            }
        }
        return [string[]]$teamsDLs.Values
    }
    
    $owner = $records[0].Owner
    $team  = $records[0].TeamDL
    $manager = $records[0].Manager
    $zoneID = [mm]::ref2ID($records[0].dnsZoneRef)

    $managerMail = [adUtils]::getEmailForUser($manager)
    $ownerMail = [adUtils]::getEmailForUser($owner)

    $recordNames = [string[]]($records | ForEach-Object { return $_.name })
    $filter = if ($records.Count -eq 1) { "name=$($recordNames[0])"} else { "name%20in%20($($recordNames -join ','))"} # just filter by name, and assume it's one record with each name most of the time

    if (!$ownerMail) {
        log "Owner $owner not found in AD. Using manager $manager ($managerMail) registered on record"
    }

    try {
        $tokens = @{
           owner = $owner
           manager = $manager
           team = $team
           zoneID = $zoneID
           application = $records[0].Application
           expiryDate= [mm]::getDateAsString($records[0].expiryDate)
           deleteDate= [mm]::getDateAsString($records[0].deleteDate)
           records = [htmlUtils]::getRecordListAsHTMLListItems($records)
           filter = $filter
        }

        $templateInfo = getTemplateInfo $template

        $to = if ($ownerMail) { @($ownerMail) } elseif ($managerMail) { @($managerMail) }
        $ccAddresses = @()
        if ([mm]::notEmpty($templateInfo.cc)){
            $ccAddresses += $templateInfo.cc
        }
        if ([mm]::notEmpty($team)) {
            $ccAddresses += [adUtils]::getGroup($team).mail
        }

        if ([mm]::isEmpty($to)){
            $to = $ccAddresses
            $ccAddresses = @()
            log "No email found for neither Owner or Manager. Sending to TeamDL $team instead: " + ($to -join ', ')
        }  
        
        if ($managerMail -and $ownerMail){
            # add manager to cc as long as we didn't have to resort to using that for the to address when no Owner mail could be found
            log "Adding manager $manager ($managerMail) to cc"
            $ccAddresses += $managerMail
        }

        $cc = $ccAddresses -join ', '

        log "Sending email to $($to -join ', ') with $cc on cc using template $($templateInfo.file) and subject: $($templateInfo.subject)" 
        [mmEmail]::prepareAndSend($to, $templateInfo.subject, $templateInfo.file, $cc, $Global:mmConfig.email.bcc, $tokens)
    } catch {
        logException 'while sending email'
    }
}

function getGroupedRecorsdNearExpiry([hashtable[]]$records, [int]$daysUntilExpiry) {    
    $matchingRecs = @()
   
    foreach ($record in $records){
        # We'll batch together records with the same properties for Owner, Team, Application, and Date (expiry and delete date)
        $record.batchString = "{0}:{1}:{2}:{3}" -f $props.Owner,$props.TeamDL,$props.Application,$props.Date

        $threshold = $record.expiryDate.AddDays(-$daysUntilExpiry)
        if ($today -ge $threshold){
            $matchingRecs += $record
        }
    }
    return $matchingRecs | groupBy -property 'batchString'   
}
function handleExpiredRecords([hashtable[]]$records){

    function handleRecordsPastGracePeriod([hashtable[]]$recordBatch){
        # records are expired and past their grace period
        # daily notifications have been sent thereof several times but now it's time to say goodbye and delete these records
        $refs = [string[]]($recordBatch | ForEach-Object { $_.ref })
        $api.call('RemoveObjects', @{objType='DNSRecord'; objRefs=$refs; saveComment = "Attestation: Removed because Owner/TeamDL took no action and let record expire."})
        sendEmailNotification $recordBatch -template ([Templates]::deleted)
    }

    # we'll assume all the records here have State=Expired
    try {
        $groupedRecords = getGroupedRecorsdNearExpiry -records $records -daysUntilExpiry 0
        foreach ($recordBatch in $groupedRecords.Values) {
            # we know each batch has the same Owner, App, and expiration date. 
            # We can simply peek at the first one to know if this batch is within the grace period or not
            if ($today -lt $recordBatch[0].deleteDate) {
                # send a warning that the records are about to be deleted 
                sendEmailNotification $recordBatch -template ([Templates]::expired)
            } else {
                handleRecordsPastGracePeriod $recordBatch
            }
        }
    } catch {
        logException 'while handling expired records'
    }
}

function handleRecordsFlaggedForDeletion($records){
    # remove all records that have been flagged for deletion ("I don't want to keep it") at the expiry date
    # TODO: Ask ebay when the records should be deleted when user flags records for deletion (immediately, imminent, at expiry date, or at deletion date)
    # we'll assume all the records on input here have State=Delete
    try {
        $groupedRecords = getGroupedRecorsdNearExpiry -records $records -daysUntilExpiry 0
        foreach ($recordBatch in $groupedRecords.Values) {
            $refs = [string[]]($recordBatch | ForEach-Object { $_.ref })
            $api.call('RemoveObjects', @{objType='DNSRecord'; objRefs=$refs; saveComment = "Attestation: Removed because Owner/Team chose to not keep the record, and flagged it for Removal."})
            sendEmailNotification $recordBatch -template ([Templates]::deleteFlag)
        }
    } catch {
        logException 'while handling expired records'
    }
}


function handleRecordsAboutToExpire([string]$state, [hashtable[]]$recordsAtState, [int]$daysUntilExpiry, [string]$newState, [string]$saveCommentSuffix){
    try {
        log "Found $($recordsAtState.Count) records with State=$state"
        
        $groupedRecords = getGroupedRecorsdNearExpiry $recordsAtState $daysUntilExpiry
        $c = 0
        foreach ($recordsBatch in $groupedRecords.Values) {
            # - send notification to Owner with TeamDL(s) in cc
            # - change state of record to the new State, such as Expiring*

            log "Found $($recordsBatch.Count) registered to $($rec.Owner) transitioning from state $state -> $newState. Notifying Owner and changing state"
            $c += $recordsBatch.Count

            sendEmailNotification -records $recordsBatch -template ([Templates]::aboutToExpire)
            foreach ($rec in $recordsBatch){
                $prefix = if ($newState -eq $kRecState.expired ) {"Record has Expired."} else { "Record is about to Expire."}
                $api.setProperties($rec.ref, 'DNSRecord', @{State=$newState}, "$prefix Owner $($rec.Owner) notified to renew. $saveCommentSuffix");
            }
        }
        log "Processed $c records transitioning from state $state -> $newState"
    } catch {
        logException "while handling records about to expire, state transitioning from $state -> $newState "
    }
}

function augmentRecords([hashtable[]]$records,[hashtable]$zone){
    # augment the records with zone and State so we can group them and process them further
    $zoneExpiryTime = [int]([mm]::getPropertyMap($zone.customProperties).'Record expiry time') ## this is in days

    if ($zoneExpiryTime -le 0){
        log "ASSERT: Zone expiry time $zoneExpiryTime expected to be larger than 0"
        return
    }

    foreach ($record in $records){
        $record.zone = $zone
        $record.fqdn = if ($record.name.EndsWith('.')) { $record.name } else { "$($record.name).$($zone.name)" }

        $props = [mm]::getPropertyMap($record.customProperties)
        $record.State = $props.State
        $record.Date = $props.Date
        $record.Owner = $props.Owner
        $record.Manager = $props.Manager
        $record.Application = $props.Application
        $record.TeamDL = $props.TeamDL
        $record.expiryDate = (get-date $record.Date).AddDays($zoneExpiryTime)
        $record.deleteDate = $record.expiryDate.AddDays($expiredGracePeriodDays)
    }
    return $records
}

$zones = $api.call('GetDNSZones', @{filter='type=Master "Record expiry time">0'}).dnsZones
log "Found $($zones.Count) zones with record expiry time. Finding records that are about to expire"

# we do this on a zone by zone basis
foreach ($zone in $zones){
    try {
        log "processing zone $($zone.name)"
        # find all records with state OK or Expiring and then process these
        $filter = "State in ({0},{1},{2},{3},{4},{5})" -f $kRecState.ok,$kRecState.expiring1,$kRecState.expiring2,$kRecState.expiring3,$kRecState.expired,$kRecState.delete
        $records = $api.call('GetDNSRecords',@{filter=$filter; dnsZoneRef=$zone.ref}).dnsRecords
        $records = augmentRecords -records $records -zone $zone

        $recordsByState = $records | groupBy 'State'

        # first, send warnings and transition state for records about to expire
        handleRecordsAboutToExpire $kRecState.ok          $recordsByState[$kRecState.ok]        $expiryWarning1Days $kRecState.expiring1 "First notification"
        handleRecordsAboutToExpire $kRecState.expiring1   $recordsByState[$kRecState.expiring1] $expiryWarning2Days $kRecState.expiring2 "Second notification"
        handleRecordsAboutToExpire $kRecState.expiring2   $recordsByState[$kRecState.expiring2] $expiryWarning3Days $kRecState.expiring3 "Third notification"
        handleRecordsAboutToExpire $kRecState.expiring3   $recordsByState[$kRecState.expiring3] $expiryWarning4Days $kRecState.expired   "Fourth and final notification"

        # then deal with expired records - we want to send a notification daily, until they're deleted after 7 days
        # we'll assume here that this script is NOT run more often than daily - because it'll send a notification every time it's run for all expired records
        handleExpiredRecords -records $recordsByState[$kRecState.expired]

        # and then we'll deal with records that had been flagged for Deletion by the user
        handleRecordsFlaggedForDeletion -records $recordsByState[$kRecState.delete]
    } catch {
        logException "while processing zone $($zone.name)"
    }

}
