﻿# version 2022.1
[CmdLetBinding()]
param(
	$username,
	$password,
	[switch]$logToConsole,
	[switch]$forceProcessing,
	[string]$inputFile
)

try {
	cd $PSScriptRoot
	$logFileName = 'workflow_submit.{0}.log' -f (get-date -Format 'yyyy-MM')
	. ./utils.ps1 -logFileName $logFileName -logToConsole:$logToConsole
	$api = [mm]::getAPI($username, $password)
} catch {
	Write-Output ('Got exception while initializing: ' + $_.Exception.GetBaseException().Message)
	exit 1
}

. ./workflow_common.ps1
$newProperties = @{}


function determineWorkflow($changeRequest) {
	$ownershipOnlyChanges = @()
    $zonesMap = @{} # ref -> workflow, group
    foreach ($change in $changeRequest.dnsRecordChanges){
        $zoneRef = ?? $change.dnsRecordAfter.dnsZoneRef $change.dnsRecordBefore.dnsZoneRef
        if (!$zonesMap.ContainsKey($zoneRef)){
            $propMap = $api.getCustomProperties($zoneRef, 'DNSZone', $false)
			$wf = $propMap.Workflow
			$zoneGroup = $propMap[$zoneProps.group]
			if ([mm]::isEmpty($wf) -or ($wf -eq $kWorkflowMap.specificGroup -and [mm]::isEmpty($zoneGroup)) ){
				# workflow is not enabled on this zone, either no workflow, or Predefined approval without a group
				returnErrorToCentral "Workflow has not been enabled for records in DNS Zone $($propMap.name) Submitting changes to records in this zone is not possible."
			}
            $zonesMap[$zoneRef] = "$wf;$zoneGroup"
        }
		if ($change.type -eq $kChangeType.mod -and [mm]::compareRecords($change.dnsRecordBefore, $change.dnsRecordAfter)){
			# the record isn't changing, only it's properties (we're just going to assume they are since the record ain't)
			$ownershipOnlyChanges += $change
		}
    }

	if ($ownershipOnlyChanges.Count -and $ownershipOnlyChanges.Count -eq $changeRequest.dnsRecordChanges.Count){
		# None of the records in the request are changing, just their properties (presumably)
		# This does not require the same sort of approval as actual modifications, and has a different workflow
		log 'Detected an Ownership update request'
		return $kWorkflowMap.ownership,"" # approval group TBD
	}

    $workflows = [string[]]($zonesMap.Values | Select-Object -Unique)
    if ($workflows.count -eq 1){
		$workflow, $approvalGroup = $workflows[0].split(';')
		return $workflow,$approvalGroup
    }
    returnErrorToCentral 'The records in the request belong to DNS Zones with different approval workflows. Please submit the changes separately.' 
}

function verifyRecordOwnership($changeRequest, [string]$fallbackTeamDL){
	
	function verifyRecordBefore($change, $changeType){
		# When removing or modifying records, the requester must be a member of the TeamDL on the record if specified
		# when not specified or group cannot be found in AD we'll just allow the requester to make the change
		# only allowing the Owner to make the change in this case would be too restrictive

		$before = $change.dnsRecordBefore
		if ([mm]::isEmpty($before)) {
			return
		}
		
		$recName = $before.name
		$props = [mm]::getPropertyMap($before.customProperties)
		$newProps = [mm]::getPropertyMap($change.dnsRecordAfter.customProperties)
		$groupInfo = if ([mm]::notEmpty($props.TeamDL)) { [adUtils]::getGroup($props.TeamDL) } else { $null }

		if ($requester -eq $props.Owner){
			log "Requester $requester is the registered Owner for record $recName, and is allowed to $changeType his own record"
			return $groupInfo # possibly stale
		}
		if ($requester -eq $props.Manager){
			log "Requester $requester is the registered Manager for record $recName, and is allowed to $changeType this record on behalf of the Owner ($($props.Owner))"
			return $groupInfo # possibly stale
		}
		
		if (!$groupInfo) {
			# what to do if the record is not registered to a TeamDL or that TeamDL is stale?
			# should we _only_ allow the owner/manager to change the record, including its ownership?
			# or do we allow anyone to request changes to it, as long as the owner stays the same (and a new TeamDL is provided that the Owner is a member of)
			# We take route2 because the first option is a bit too restrictive - but this is easy to change or to have configurable
			$isModifyingOwner = $newProps -and [mm]::notEmpty($props.Owner) -and $newProps.Owner -ne $props.Owner

			if ($isModifyingOwner -and [adUtils]::verifyUserByName($props.Owner)){
				returnErrorToCentral "You cannot change Owner $($props.Owner) for record $recName to $($newProps.Owner), because $($props.Owner) is still active and the TeamDL $($props.TeamDL) needs to be updated to a valid TeamDL first."
			}

			log "TeamDL not registered or valid on record  $($before.name), and Owner is either unset or not being modified. Allowing requester to $changeType"
			return 
		}

		if ([adUtils]::isUserMemberInGroupByDN($changeRequest.requester, $groupInfo.dn) ){
			log "$requester is member of registered (active) TeamDL $($props.TeamDL), and is allowed to request change to the record $recName"
			return $groupInfo
		} else {
			returnErrorToCentral "You cannot request to $changeType the DNS Record $($before.name) You are not a member of the AD group $($props.TeamDL) that it is registered to."
		}
	}

	function verifyTeamDL($teamDL, $owner, $recordTeamSuffix){
		# verify record Team
		$groupInfo = [adUtils]::getGroup($teamDL)
		if ($null -eq $groupInfo){
			returnErrorToCentral "No group could be found in Active Directory with the name `"$teamDL`" $recordTeamSuffix"
		}
		
		# owner must always be a member in the group
		if (![adUtils]::isUserMemberInGroupByDN($owner, $groupInfo.dn)){
			returnErrorToCentral "Cannot $changeType record because $owner is not a member of the AD group $($teamDL) $recordTeamSuffix"
		}

		# the teamDL must have a mailbox, and have at least two active members
		if ([mm]::isEmpty($groupInfo.mail)){
			returnErrorToCentral "The group `"$teamDL`" $recordTeamSuffix is not configured with email in Active Directory. Please specify an email configured AD group instead."
		}
		$minGroupMembers = ?? $Global:mmConfig.AD.minTeamDLMembers 2
		if ([adUtils]::getUsersInGroupByDN($groupInfo.dn, @(), $true).Count -lt $minGroupMembers){
			returnErrorToCentral "The TeamDL `"$teamDL`" $recordTeamSuffix has too few users as members in Active Directory. Please enter a different TeamDL with $minGroupMembers or more members."
		}

		return $groupInfo
	}

	function verifyRecordAfter($change, $changeType){
		# for both add and modify, on dnsRecordAfter
		# - TeamDL must be specified and valid, or the fallback must be valid
		# - Owner must be specified and valid and be member of TeamDL
		# - requester must be member of TeamDL too
		$rec = $change.dnsRecordAfter
		if ([mm]::isEmpty($rec)){
			return
		}

		$props = [mm]::getPropertyMap($rec.customProperties)
		$owner = $props.Owner
		$teamDL = ?? $props.TeamDL $fallbackTeamDL
		$recName = $rec.name
		
		# verify record Owner (which ebay has as a required field) is valid, if different than requester which we know is valid
		if ($owner -ne $requester -and ![adUtils]::verifyUserByName($owner)){
			returnErrorToCentral "Owner `"$owner`" specified for record $recName could not be found in Active Directory, or is not active" 
		}

		if ([mm]::isEmpty($teamDL)) {
			# no fallback, no team - ask for TeamDL on record if predefined approval, but for Group if peer review (because it's easier for the requester rather than going back and editing one/all of the changes)
			if ($workflowType -eq $kWorkflowMap.specificGroup){
				returnErrorToCentral "TeamDL must be specified for record $recName"
			} else {
				returnErrorToCentral "Please specify an AD group you are a member of in `"Group`" and try again"
			}
		}
		
		$recordTeamSuffix = if ([mm]::notEmpty($props.TeamDL)) {"specified as TeamDL for the record $recName"} else {""}

		# verify record Team
		$groupInfo = verifyTeamDL $teamDL $owner $recordTeamSuffix

		# requester must be in the TeamDL when adding records, if he's not the owner
		# Requester does not need to be in the new TeamDL when modifying a record - verifyRecordBefore ensures existing owners are requesting the modification
		# this allows existing owners to assign records to different teams (or anyone to re-assign to a new team when the existing team is stale)
		if ($change.type -eq $kChangeType.add -and $owner -ne $requester -and ![adUtils]::isUserMemberInGroupByDN($requester, $groupInfo.dn)){
			returnErrorToCentral "Cannot $changeType record because you are not a member of the AD group $($teamDL) $recordTeamSuffix"
		}
		
		return $groupInfo
	}

	$requester = $changeRequest.requester
	$teamDLs = @{}
	foreach ($change in $changeRequest.dnsRecordChanges){
		$changeType = $change.type.replace('DNSRecord','')
		if ($change.type -eq $kChangeType.add){
			$afterProps = [mm]::getPropertyMap($change.dnsRecordAfter.customProperties)
			$team = ?? $afterProps.TeamDL $fallbackTeamDL
			log "$requester is requesting to $changeType record $($change.dnsRecordAfter.name) on behalf of $($afterProps.Owner) in $team"
		} else {
			$beforeProps = [mm]::getPropertyMap($change.dnsRecordBefore.customProperties)
			log "$requester is requesting to $changeType record $($change.dnsRecordBefore.name) that is registered to $($beforeProps.Owner) in $($beforeProps.TeamDL)"
		}

		$beforeGroupInfo = verifyRecordBefore $change $changeType
		$afterGroupInfo = verifyRecordAfter $change $changeType

		$approvingTeamInfo = ?? $beforeGroupInfo $afterGroupInfo # if the existing record has a valid team, that team approves the modification, including if to assign the record to another team
		if ($null -eq $approvingTeamInfo -and [mm]::notEmpty($fallbackTeamDL)) {
			$approvingTeamInfo = verifyTeamDL $fallbackTeamDL $requester ""
		}
		if ($approvingTeamInfo){
			$teamDLs[$approvingTeamInfo.samAccountName] = $approvingTeamInfo
		}
	}
	return $teamDLs
}

function handlePeerReviewRequest($changeRequest, [string]$groupStr){
	function verifyChanges($fallbackTeam){
		# first verify a couple of naming policies enforced by ebay
		# - no apex
		# - no multi.labels
		foreach ($change in $changeRequest.dnsRecordChanges){
			if ($change.type -eq $kChangeType.add){
				$zonename = $api.call('GetDNSZone', @{dnsZoneRef = $change.dnsRecordAfter.dnsZoneRef}).dnsZone.name
				if ($change.dnsRecordAfter.name -eq $zonename){
					returnErrorToCentral "$zonename is not an allowed record name: It is an apex record for the zone $zonename"
				}
	
				$pqdn = $change.dnsRecordAfter.name.replace('.'+ $zonename, '')
				if ($pqdn.split('.').Count -gt 1){
					returnErrorToCentral "$pqdn is not an allowed hostname: Only single label hostnames are allowed."
				}
			}
		}

		$teamDLs = [hashtable](verifyRecordOwnership -changeRequest $changeRequest -fallbackTeamDL $fallbackTeam)

		# we want the same TeamDL to be used on all records where specified (and for it to be consistent with the Request Group)
		if ($teamDLs.Count -gt 1){
			# requester is submitting records for multiple Teams at the same time 
			# We don't allow this in peer review requests because the changes must be approved by a single Team
			returnErrorToCentral "Cannot submit a Peer Review request with records belonging to multiple different teams ($($teamDLs.Keys -join ', '))."
		}

		if ($teamDLs.Count -eq 1){
			return $teamDLs.Values[0] # the group info for the TeamDL on the records
		}
		return $null # no TeamDL on records, nor a fallback (group on the request) supplied
	}

	$teamGroupInfo = verifyChanges -fallbackTeam $groupStr
	
	if (!$teamGroupInfo){
		# this means team was neither specified on records nor was there a fallback specified
		if ([mm]::isEmpty($groupStr)){
			returnErrorToCentral "Please enter the name of the distribution group for your Team in `"Group`""
		} else {
			log ("ASSERT: Group supplied but no team group info found")
			returnErrorToCentral "Please specify a TeamDL for the records in the request" # paranoia: should never happen, but by returning here we can be certain that teamGroupInfo is non-null
		}
	}
	

	if ([mm]::notEmpty($groupStr)) {
		# if user supplied a group we need to check that it's valid and matches the group for the record TeamDL
		$adGroup = [adUtils]::getGroup($groupStr)
		if (!$adGroup){
			returnErrorToCentral "No AD group found that matches `"$groupStr`""
		}

		# enforce that the record team group and the supplied request groups are one and the same
		if ($teamGroupInfo.samAccountName -ne $adGroup.samAccountName) {
			returnErrorToCentral "Unable to submit request because the TeamDL $($teamGroupInfo.samAccountName) that was specified on the records in the request is not the same as the group $($adGroup.samAccountName) specified in field `"Group`"."
		}
		
		# we don't need to check membership, that has already been done as part of validating the teamDL on the changes, which we now know is the same as the Group
	}

	if (!$teamGroupInfo.mail){
		returnErrorToCentral "The AD group $($teamGroupInfo.samAccountName) is not configured with an email address. Please enter a distribution group with email configured."
	}

	return $teamGroupInfo
}

function handlePredefinedApprovalRequest($changeRequest, $approvalGroup) {
	# Check all the record TeamDLs on the records in the request
	# - All records must specify valid TeamDLs and Owners on dnsRecordAfter 
	# - the Owner must be a valid user that is a member of the TeamDL on each record
	# - Only existing record owner or members of TeamDL can modify/delete existing records 
	$null = verifyRecordOwnership -changeRequest $changeRequest # no fallback

	# let's fetch and verify the approval group specified on the zone. If it isn't valid, it's misconfigured
	$groupInfo = [adUtils]::getGroup($approvalGroup)
	if (!$groupInfo){
		returnErrorToCentral "Approval Group $approvalGroup configured on the DNS Zone(s) in the request cannot be found in Active Directory"
	}
	if ([mm]::isEmpty($groupInfo.mail)){
		returnErrorToCentral "Approval Group $approvalGroup configured on the DNS Zone(s) is not configured with an email address and cannot be used as Approval Group."
	}
	
	# all is good ... let's return the group to store on the request
	return $groupInfo
}

function getApprovers([hashtable]$groupInfo, [boolean]$includeManager){

	# start by adding the members of the group as approvers
	$approvers = @()
	$groupMembers = [adUtils]::getUsersInGroupByDN($groupInfo.dn, @(), $true)
	foreach ($member in $groupMembers){
		$memberName = $member.'msDS-PrincipalName'
		if ($memberName -ne $requester -and $memberName -notin $approvers){
			$approvers += $memberName
		}
	}

	if (!$approvers.Count){
		returnErrorToCentral "The group $($groupInfo.samAccountName) has too few users as members. Please enter a different distribution group with more members."
	}

	# add the manager if not already in the approvers list
	if ($includeManager){
		$manager = [adUtils]::getManagerForUser($requester)
		if ($manager -and $manager.name){
			if ($manager.name -notin $approvers){
				log ('Manager {0} not part of group. Adding to approvers list' -f $manager.name)
				$approvers += $manager.name
			}
		} else {
			log ("No manager registered for requester $requester in AD!")
		}
	}

	# If there are that many approvers that we can't save it in the database, just store the first 890 characters.
	# That should be enough for 100 approvers or so
	$approversStr = $approvers -join "`n"
	if ($approversStr.Length -gt 890){
		$approversStr = $approversStr.Substring(0,890)
	}

	return $approversStr
}

function verifyPropertyChangeRequest([hashtable]$changeRequest) {

	$automatedChanges = @()
	foreach ($change in $changeRequest.customPropertyChanges){
		if ($change.ref.startsWith('DNSRecords/')) {
			$updatedProps  = [mm]::getPropertyMap($change.customPropertiesAfter)
			if (!$updatedProps.ContainsKey('State')){
				# State is not being updated. Guess we don't care about this change then
				log "State is not being supplied for record $($change.ref). Not verifying nor automating"
				continue
			}

			if ($updatedProps.Count -ne 1){
				# something more than State is being changed. We don't automate unless State is the only thing changing
				log "Other properties than State are being changed too here. Not verifying or automating"
				continue
			}

			$currentProps = $api.getCustomProperties($change.ref, 'DNSRecord')
			if ($updatedProps.State -eq $currentProps.State) {
				log "State is not being changed for record $($change.ref). Not verifying nor automating"
				continue
			}

			# we'll allow the change if there's no owner(ship)
			# but if there is, the requester must be the owner or the manager, or a member of the PDL
			if ([mm]::isEmpty($currentProps.Owner) `
				-or $changeRequest.requester -eq $currentProps.Owner `
				-or $changeRequest.requester -eq $currentProps.Manager `
				-or [adUtils]::isUserMemberInGroup($changeRequest.requester, $currentProps.TeamDL)
			){
				$automatedChanges += $change
			} else {
				# possibly need to be more specific here? Like if the TeamDL cannot be found in AD ?
				returnErrorToCentral "You are not the Owner of the record ($($currentProps.Owner)) or a member of the TeamDL $($currentProps.TeamDL)"
			}
		}
	}
	if ($automatedChanges.Count) {
		log "Approving/automating $($automatedChanges.Count) state-change changes that the requester owns or is on registered team"
		$api.callAsync('ApproveChanges', @{changeRequestRef = $changeRequest.ref; customPropertyChanges=$automatedChanges})
	}
}

try {
	if ([mm]::notEmpty($inputFile)){
		logv2 "Reading input from $inputFile"
		$xmlStr = Get-Content -Raw $inputFile
	} else {
		$xmlStr = [Console]::In.ReadToEnd()
		[mm]::storeDebugBlob($xmlStr, 'submit', 'xml')
		logv2 ("Input:" + $xmlStr)
	}


	$xml = [xml]($xmlStr)
	$requester = [mm]::cleanXmlString(($xml.selectSingleNode("externalScriptParameters")).getAttribute('userName'))
	$cfs = $xml.selectSingleNode("externalScriptParameters/customFields")
	$groupStr = [mm]::cleanXmlString($cfs.selectSingleNode("customField[@customFieldName='Group']").value)
    $workflowProp = [mm]::cleanXmlString($cfs.selectSingleNode("customField[@customFieldName='Workflow']").value)
	$requestID = [int]$xml.externalScriptParameters.object.id
    $requestRef = [mm]::ID2ObjRef($requestID, 28)

	# new requests don't have "Workflow" property set - should be read only and required field on ChangeRequests
	# ideally we wouldn't need workflow to be set to anything, but then this script wont get triggered when no properties are supplied
	# therefore Workflow needs to be a required field, so this script is always triggered
	if ($workflowProp -ne '--' -and !$forceProcessing){
		log "Request already processed (Workflow set to $workflowProp)"
		# this should either only happen if an API user doesn't supply '--' as Workflow, or someone is trying to change properties for an already submitted request
		# we'll simply not allow that, so we can enforce API users to supply -- as the UI does
		if ($requester -eq 'ScriptRunner' -or $requester -eq 'administrator') {
			# ok we make an exception for ScriptRunner and administrator
			log "Allowing $requester to make changes to request properties"
			exit 0
		} else {
			returnErrorToCentral "Workflow must be set to '--'"
		}
	}

	log ("Processing request {0} submitted by {1}" -f $requestID, $requester)

    # first we need to dermine which kind of workflow there is from the zone(s) in the request
    #
    # Predefined approval
    #    - "Group" should be ignored. Follows the Approval group for the zone(s), which must be the same for all changes
    #    - requester is not (necessarily) a member of "Group"
    #    - Ownership on records in request is validated, Owners must be members in TeamsDLs and so forth
	#	 - Records are allowed to belong to multiple Owners and/or Teams
    #
    # Peer review
    #    - Group here is the TeamDL for the records. 
	#		- If the user supplies a Group it must be the same as the mutual TeamDL on the records. 
	#		- If not supplied automatically use the TeamDL on the records
	#    	- Requester must be a member of the TeamDL/Group 
    #    - Only a single TeamDL is allowed. Error thrown if different Teams on different records in the request
	# 	 - Ownership on records validated, such that Owners are members of TeamDLs etc
  

    
	log 'Fetching request'
	$changeRequest = $api.call('GetChangeRequest', @{changeRequestRef = $requestRef}).changeRequest

	if ($changeRequest.objType -eq 'DNSRecord'){

		$workflowType, $approvalGroup = determineWorkflow -changeRequest $changeRequest
		$newProperties.Workflow = $workflowType
		
		$approvalGroupInfo = $null
		$managerCanApprove = $true
		if ($workflowType -eq $kWorkflowMap.peer){
			$approvalGroupInfo = handlePeerReviewRequest -changeRequest $changeRequest -groupStr $groupStr
		} elseif ($workflowType -eq $kWorkflowMap.ownership){
			# actually, peer review also applies for ownership requests if for some reason not automated
			# specific group (infosec) should not be the backup to approve these 
			$approvalGroupInfo = handlePeerReviewRequest -changeRequest $changeRequest -groupStr $groupStr
		} elseif ($workflowType -eq $kWorkflowMap.specificGroup){
			$approvalGroupInfo = handlePredefinedApprovalRequest -changeRequest $changeRequest -approvalGroup $approvalGroup
			$managerCanApprove = $false # only members of specific group can approve
		} else {
			returnErrorToCentral "Unknown Workflow `"$workflowType`" configured on a DNS Zone"
		}

		if (!$approvalGroupInfo){
			log "ASSERT: group should already be valid"
			returnErrorToCentral "Please specify a valid Group"
		}

		if ($approvalGroupInfo.samAccountName -ne $groupStr) {
			$newProperties.Group = $approvalGroupInfo.samAccountName
		}

		$newProperties.Approvers = getApprovers -groupInfo $approvalGroupInfo -includeManager $managerCanApprove
	}

	if ($changeRequest.objType -eq 'CustomProperties'){
		# renewing or expiring records 
		verifyPropertyChangeRequest -changeRequest $changeRequest
		$newProperties.Workflow = $kWorkflowMap.renewal
	}

	returnSuccessToCentral $newProperties
} catch {
	logException "while processing input xml"
	log $xmlStr # repeated if in higher verbosity
	returnErrorToCentral "Unhandled exception"
} finally {
	$api.logout()
}
